/******************************************************************************
  This file contains routines that can be bound to a Postgres backend and
  called by the backend in the process of processing queries.  The calling
  format for these routines is dictated by Postgres architecture.
******************************************************************************/

#include <stdio.h>
#include <float.h>
#include <string.h>

#include "postgres.h"
#include "access/gist.h"
#include "access/itup.h"
#include "access/rtree.h"
#include "utils/elog.h"
#include "utils/palloc.h"
#include "utils/array.h"
#include "utils/builtins.h"
#include "storage/bufpage.h"


#define MAXNUMRANGE 5
#define CONSTLENKEY

#define max(a,b)        ((a) >  (b) ? (a) : (b))
#define min(a,b)        ((a) <= (b) ? (a) : (b))
#define abs(a)          ((a) <  (0) ? (-a) : (a))

#define ARRPTR(x)  ( (int4 *) ARR_DATA_PTR(x) )
#define ARRSIZE(x)  getNitems( ARR_NDIM(x), ARR_DIMS(x))

#define SORT(a,b) ( ( b<20 ) ? qsort(a,b,sizeof(int), sort_intarray) : mergesort(a,b,sizeof(int), sort_intarray) )

#define NDIM 1
#define ARRISNULL(x) ( ! ( x && ARR_NDIM(x) == NDIM && ARRSIZE(x) ) )
#define PREPAREARR \
	na = ARRSIZE( a ); nb = ARRSIZE( b );\
	da = ARRPTR( a );\
	db = ARRPTR( b );\
	SORT( (void*)da, na );\
	SORT( (void*)db, nb );

#define GIST_DEBUG
#define GIST_QUERY_DEBUG 
/*
static void printarr ( ArrayType * a, int num ) {
	char bbb[16384];
	char *cur;
	int l;
	int *d;
	d = ARRPTR( a );
	*bbb = '\0';
	cur = bbb;
	for(l=0; l<min( num, ARRSIZE( a ));l++) {
		sprintf(cur,"%d ", d[l] );
		cur = strchr( cur, '\0' ) ;
	}
	elog(NOTICE, "\t\t%s", bbb);
}
*/

/*
** usefull function
*/
int sort_intarray( const void *a, const void *b );
ArrayType * new_intArrayType( int num );
ArrayType * copy_intArrayType( ArrayType * a );
ArrayType * resize_intArrayType( ArrayType * a, int num );
int bytesize_intarray (ArrayType * a);
int internal_size( int *a, int len );

/* 
** GiST support methods
*/
bool             g_int_consistent(GISTENTRY *entry, ArrayType *query, StrategyNumber strategy);
GISTENTRY *      g_int_compress(GISTENTRY *entry);
GISTENTRY *      g_int_decompress(GISTENTRY *entry);
float *          g_int_penalty(GISTENTRY *origentry, GISTENTRY *newentry, float *result);
GIST_SPLITVEC *  g_int_picksplit(bytea *entryvec, GIST_SPLITVEC *v);
bool             g_int_leaf_consistent(ArrayType *key, ArrayType *query, StrategyNumber strategy);
bool             g_int_internal_consistent(ArrayType *key, ArrayType *query, StrategyNumber strategy);
ArrayType *            g_int_union(bytea *entryvec, int *sizep);
bool *           g_int_same(ArrayType *b1, ArrayType *b2, bool *result);


/*
** R-tree suport functions
*/
bool     _int_different(ArrayType *a, ArrayType *b);
bool     _int_same(ArrayType *a, ArrayType *b);
bool     _int_contains(ArrayType *a, ArrayType *b);
bool     _int_contained(ArrayType *a, ArrayType *b);
bool     _int_overlap(ArrayType *a, ArrayType *b);
ArrayType *    _int_union(ArrayType *a, ArrayType *b);
ArrayType *    _int_inter(ArrayType *a, ArrayType *b);
void     rt__int_size(ArrayType *a, float* sz);


/*****************************************************************************
 *                         GiST functions
 *****************************************************************************/

/*
** The GiST Consistent method for _intments
** Should return false if for all data items x below entry,
** the predicate x op query == FALSE, where op is the oper
** corresponding to strategy in the pg_amop table.
*/
bool 
g_int_consistent(GISTENTRY *entry,
	       ArrayType *query,
	       StrategyNumber strategy)
{
   
    /* sort query for fast search, key is already sorted */ 
    SORT( (void*)ARRPTR( query ), ARRSIZE( query ) );    
    /*
    ** if entry is not leaf, use g_int_internal_consistent,
    ** else use g_int_leaf_consistent
    */
/*    if (GIST_LEAF(entry))
      return(g_int_leaf_consistent((ArrayType *)(entry->pred), query, strategy));
    else
*/
      return(g_int_internal_consistent((ArrayType *)(entry->pred), query, strategy));
}

/*
** The GiST Union method for _intments
** returns the minimal set that encloses all the entries in entryvec
*/
ArrayType *
g_int_union(bytea *entryvec, int *sizep)
{
    int numranges, i;
    ArrayType *out = (ArrayType *)NULL;
    ArrayType *tmp;

    numranges = (VARSIZE(entryvec) - VARHDRSZ)/sizeof(GISTENTRY); 
    tmp = (ArrayType *)(((GISTENTRY *)(VARDATA(entryvec)))[0]).pred;

#ifdef GIST_DEBUG
    elog(NOTICE, "union %d", numranges);
#endif

    for (i = 1; i < numranges; i++) {
	out = _int_union(tmp, (ArrayType *)
				 (((GISTENTRY *)(VARDATA(entryvec)))[i]).pred);
	if (i > 1) pfree(tmp);
	tmp = out;
    }

    *sizep = bytesize_intarray( out );
#ifdef GIST_DEBUG
    elog(NOTICE, "\t ENDunion %d %d", *sizep, ARRSIZE( out ) );
#endif
    return(out);
}

/*
** GiST Compress and Decompress methods
*/
GISTENTRY *
g_int_compress(GISTENTRY *entry)
{
    GISTENTRY *retval = palloc(sizeof(GISTENTRY));
    ArrayType * r = copy_intArrayType( (ArrayType *) entry->pred );
    int len = ARRSIZE(r);
    int *dr;
    int i,min,cand;

#ifdef GIST_DEBUG
	elog(NOTICE, "COMP IN: %d elems", len);
#endif

    if ( ! retval ) 
	elog(ERROR,"Can't allocate memory for compression");

#ifdef CONSTLENKEY 
    r = resize_intArrayType( r, 2*max(len, MAXNUMRANGE) );
#else
    r = resize_intArrayType( r, 2*( len ) );
#endif
   
    dr = ARRPTR( r );
 
    if ( len > 1 ) 
	SORT( (void*)dr, len );

    for(i=len-1; i>=0;i--)
    	dr[2*i] = dr[2*i+1] = dr[i];
    
    if ( len <= MAXNUMRANGE ) {
#ifdef CONSTLENKEY
	for(i=len * 2; i < MAXNUMRANGE * 2; i++)
		dr[i]=dr[2*(len-1)+1];
	len = 2*MAXNUMRANGE;
#else
	len *= 2;
#endif
    } else {
	len *= 2;
	cand = 1;
	while( len > MAXNUMRANGE * 2 ) {
		min = 0x7fffffff;
		for( i=2; i<len;i+=2 )
			if ( min > (dr[i] - dr[i-1]) ) {
				min = (dr[i] - dr[i-1]);
				cand = i;
			}
		memmove( (void*)&dr[cand-1], (void*)&dr[cand+1], (len - cand - 1)*sizeof(int) );
		len -= 2;
	}
    }
    r = resize_intArrayType(r, len );

    gistentryinit(*retval, (char *)r, entry->rel, entry->page, entry->offset, bytesize_intarray( r ), FALSE); 

#ifdef GIST_DEBUG
	elog(NOTICE, "COMP OUT: %d elems", len);
#endif
    return(retval);
}

GISTENTRY *
g_int_decompress(GISTENTRY *entry)
{
    GISTENTRY *retval;
    ArrayType * r; 
    int *dr, lenr;
    ArrayType * in = (ArrayType *) entry->pred; 
    int lenin = ARRSIZE(in);
    int *din = ARRPTR(in);
    int i,j;

#ifdef GIST_DEBUG
	elog(NOTICE, "DECOMP IN: %d leaf; %d elems", entry->leafkey, lenin);
#endif
    /* sometimes entry->leafkey == TRUE !?!?!? */
    if ( entry->leafkey ) return( entry );

    retval = palloc(sizeof(GISTENTRY));
    if ( ! retval ) 
	elog(ERROR,"Can't allocate memory for decompression");

    lenr = internal_size(din, lenin);

    r = new_intArrayType( lenr );
    dr = ARRPTR( r );

    for(i=0;i<lenin;i+=2)
	for(j=din[i]; j<=din[i+1]; j++)
		if ( (!i) || *(dr-1) != j )
			*dr++ = j;

    gistentryinit(*retval, (char *)r, entry->rel, entry->page, entry->offset, bytesize_intarray( r ), FALSE); 

#ifdef GIST_DEBUG
	elog(NOTICE, "DECOMP OUT: %d elems", lenr);
#endif
    return(retval);
}

/*
** The GiST Penalty method for _intments
*/
float *
g_int_penalty(GISTENTRY *origentry, GISTENTRY *newentry, float *result)
{
    Datum ud;
    float tmp1, tmp2;
    
    ud = (Datum)_int_union((ArrayType *)(origentry->pred), (ArrayType *)(newentry->pred));
    rt__int_size((ArrayType *)ud, &tmp1);
    rt__int_size((ArrayType *)(origentry->pred), &tmp2);
    *result = tmp1 - tmp2;
    pfree((char *)ud);

#ifdef GIST_DEBUG
    elog(NOTICE, "penalty\t%g", *result);
#endif

    return(result);
}



/*
** The GiST PickSplit method for _intments
** We use Guttman's poly time split algorithm 
*/
GIST_SPLITVEC *
g_int_picksplit(bytea *entryvec,
	      GIST_SPLITVEC *v)
{
    OffsetNumber i, j;
    ArrayType *datum_alpha, *datum_beta;
    ArrayType *datum_l, *datum_r;
    ArrayType *union_d, *union_dl, *union_dr;
    ArrayType *inter_d;
    bool firsttime;
    float size_alpha, size_beta, size_union, size_inter;
    float size_waste, waste;
    float size_l, size_r;
    int nbytes;
    OffsetNumber seed_1 = 0, seed_2 = 0;
    OffsetNumber *left, *right;
    OffsetNumber maxoff;

#ifdef GIST_DEBUG
    elog(NOTICE, "--------picksplit %d",(VARSIZE(entryvec) - VARHDRSZ)/sizeof(GISTENTRY));
#endif

    maxoff = ((VARSIZE(entryvec) - VARHDRSZ)/sizeof(GISTENTRY)) - 2;
    nbytes =  (maxoff + 2) * sizeof(OffsetNumber);
    v->spl_left = (OffsetNumber *) palloc(nbytes);
    v->spl_right = (OffsetNumber *) palloc(nbytes);
    
    firsttime = true;
    waste = 0.0;
    
    for (i = FirstOffsetNumber; i < maxoff; i = OffsetNumberNext(i)) {
	datum_alpha = (ArrayType *)(((GISTENTRY *)(VARDATA(entryvec)))[i].pred);
	for (j = OffsetNumberNext(i); j <= maxoff; j = OffsetNumberNext(j)) {
	    datum_beta = (ArrayType *)(((GISTENTRY *)(VARDATA(entryvec)))[j].pred);
	    
	    /* compute the wasted space by unioning these guys */
	    /* size_waste = size_union - size_inter; */
	    union_d = (ArrayType *)_int_union(datum_alpha, datum_beta);
	    rt__int_size(union_d, &size_union);
	    inter_d = (ArrayType *)_int_inter(datum_alpha, datum_beta);
	    rt__int_size(inter_d, &size_inter);
	    size_waste = size_union - size_inter;
	    
	    pfree(union_d);
	    
	    if (inter_d != (ArrayType *) NULL)
		pfree(inter_d);
	    
	    /*
	     *  are these a more promising split that what we've
	     *  already seen?
	     */
	    
	    if (size_waste > waste || firsttime) {
		waste = size_waste;
		seed_1 = i;
		seed_2 = j;
		firsttime = false;
	    }
	}
    }
    
    left = v->spl_left;
    v->spl_nleft = 0;
    right = v->spl_right;
    v->spl_nright = 0;
  
    datum_alpha = (ArrayType *)(((GISTENTRY *)(VARDATA(entryvec)))[seed_1].pred);
    datum_l     = copy_intArrayType( datum_alpha ); 
    rt__int_size((ArrayType *)datum_l, &size_l);
    datum_beta  = (ArrayType *)(((GISTENTRY *)(VARDATA(entryvec)))[seed_2].pred);
    datum_r     = copy_intArrayType( datum_beta  ); 
    rt__int_size((ArrayType *)datum_r, &size_r);
    
    /*
     *  Now split up the regions between the two seeds.  An important
     *  property of this split algorithm is that the split vector v
     *  has the indices of items to be split in order in its left and
     *  right vectors.  We exploit this property by doing a merge in
     *  the code that actually splits the page.
     *
     *  For efficiency, we also place the new index tuple in this loop.
     *  This is handled at the very end, when we have placed all the
     *  existing tuples and i == maxoff + 1.
     */
    
    maxoff = OffsetNumberNext(maxoff);
    for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i)) {
	
	/*
	 *  If we've already decided where to place this item, just
	 *  put it on the right list.  Otherwise, we need to figure
	 *  out which page needs the least enlargement in order to
	 *  store the item.
	 */
	
	if (i == seed_1) {
	    *left++ = i;
	    v->spl_nleft++;
	    continue;
	} else if (i == seed_2) {
	    *right++ = i;
	    v->spl_nright++;
	    continue;
	}
	
	/* okay, which page needs least enlargement? */ 
	datum_alpha = (ArrayType *)(((GISTENTRY *)(VARDATA(entryvec)))[i].pred);
	union_dl = (ArrayType *)_int_union(datum_l, datum_alpha);
	union_dr = (ArrayType *)_int_union(datum_r, datum_alpha);
	rt__int_size((ArrayType *)union_dl, &size_alpha);
	rt__int_size((ArrayType *)union_dr, &size_beta);

	/* pick which page to add it to */
	if (size_alpha - size_l < size_beta - size_r) {
	/*if ( ((float)(size_alpha - size_l))/(float)size_r < ((float)(size_beta - size_r))/(float)size_l ) {*/
	    pfree(datum_l);
	    pfree(union_dr);
	    datum_l = union_dl;
	    size_l = size_alpha;
	    *left++ = i;
	    v->spl_nleft++;
	} else {
	    pfree(datum_r);
	    pfree(union_dl);
	    datum_r = union_dr;
	    size_r = size_alpha;
	    *right++ = i;
	    v->spl_nright++;
	}
    }
/*     *left = *right = FirstOffsetNumber;	*/ /* sentinel value, see dosplit() */

/*    if ( v->spl_nleft > v->spl_nright ) { */
    if ( *(left-1) > *(right-1) ) { /* last update */ 
	*right = FirstOffsetNumber;	
	*left = InvalidOffsetNumber;
    } else {
	*left = FirstOffsetNumber;
	*right = InvalidOffsetNumber;	
    }

    v->spl_ldatum = (char *)datum_l;
    v->spl_rdatum = (char *)datum_r;

#ifdef GIST_DEBUG
    elog(NOTICE, "--------ENDpicksplit %d %d",v->spl_nleft, v->spl_nright);
#endif
    return v;
}

/*
** Equality methods
*/


bool *
g_int_same(ArrayType *b1, ArrayType *b2, bool *result)
{
  if (_int_same(b1, b2))
    *result = TRUE;
  else *result = FALSE;

  return(result);
}

/* 
** SUPPORT ROUTINES
*/
bool 
g_int_leaf_consistent(ArrayType *key,
		     ArrayType *query,
		     StrategyNumber strategy)
{
    bool retval;

#ifdef GIST_QUERY_DEBUG
  elog(NOTICE, "leaf_consistent, %d", strategy);
#endif

    switch(strategy) {
    case RTOverlapStrategyNumber:
      retval = (bool)_int_overlap(key, query);
      break;
    case RTSameStrategyNumber:
      retval = (bool)_int_same(key, query);
      break;
    case RTContainsStrategyNumber:
      retval = (bool)_int_contains(key, query);
      break;
    case RTContainedByStrategyNumber:
      retval = (bool)_int_contained(key,query);
      break;
    default:
      retval = FALSE;
    }
    return(retval);
}

bool 
g_int_internal_consistent(ArrayType *key,
			ArrayType *query,
			StrategyNumber strategy)
{
    bool retval;

#ifdef GIST_QUERY_DEBUG
  elog(NOTICE, "internal_consistent, %d", strategy);
#endif

    switch(strategy) {
    case RTOverlapStrategyNumber:
      retval = (bool)_int_overlap(key, query);
      break;
    case RTSameStrategyNumber:
    case RTContainsStrategyNumber:
      retval = (bool)_int_contains(key, query);
      break;
    case RTContainedByStrategyNumber:
      retval = (bool)_int_overlap(key, query);
      break;
    default:
      retval = FALSE;
    }
    return(retval);
}

bool 
_int_contains ( ArrayType *a, ArrayType *b ) {
	int na , nb ;
	int i,j, n;
	int *da, *db;

	if ( ARRISNULL( a ) || ARRISNULL( b ) ) return FALSE;
	PREPAREARR;

#ifdef GIST_DEBUG
    elog(NOTICE, "contains %d %d", na, nb);
#endif

	i = j = n = 0;
	while( i<na && j<nb )
		if ( da[i] < db[j] )
			i++;
		else if ( da[i] == db[j] ) {
			n++; i++; j++;
		} else 
			j++;
	
	return ( n == nb ) ? TRUE : FALSE;
}

bool
_int_contained(ArrayType *a, ArrayType *b)
{
  return ( _int_contains(b, a) );
}

/*****************************************************************************
 * Operator class for R-tree indexing
 *****************************************************************************/

bool
_int_different(ArrayType *a, ArrayType *b)
{
  return ( !_int_same( a, b ) );
}
bool 

_int_same ( ArrayType *a, ArrayType *b ) {
        int na , nb ;
        int n; 
        int *da, *db;
	bool anull = ARRISNULL( a );
	bool bnull = ARRISNULL( b );

	if ( anull || bnull ) 
		return ( anull && bnull ) ? TRUE : FALSE; 
			

        PREPAREARR;

        if ( na != nb ) return FALSE;

        n = 0;
        for(n=0; n<na; n++)
                if ( da[n] != db[n] )
                        return FALSE;

        return TRUE; 
}

/*  _int_overlap -- does a overlap b?
 */
bool 
_int_overlap ( ArrayType *a, ArrayType *b ) {
	int na , nb ;
	int i,j;
	int *da, *db;

	if ( ARRISNULL( a ) || ARRISNULL( b ) ) return FALSE;
	PREPAREARR;

#ifdef GIST_DEBUG
    elog(NOTICE, "g_int_overlap");
#endif

	i = j = 0;
	while( i<na && j<nb )
		if ( da[i] < db[j] )
			i++;
		else if ( da[i] == db[j] )
			return TRUE; 
		else 
			j++;
	
	return FALSE;
}

ArrayType * 
_int_union ( ArrayType *a, ArrayType *b ) {
	ArrayType * r = NULL;
	int na , nb;
	int *da, *db, *dr, *tmp;
	int i,j;

	if ( ARRISNULL( a ) ) r = copy_intArrayType( b ); 
	if ( ARRISNULL( b ) ) r = copy_intArrayType( a ); 

	if ( r ) { 
		dr = ARRPTR( r );
		SORT( (void*)r, ARRSIZE( r ) );
	} else {
		PREPAREARR;

		r = new_intArrayType( na + nb ); 
		dr = ARRPTR( r );

		/* union */	
		i = j = 0;
		while( i<na && j<nb ) 
			if ( da[i] < db[j] )
				*dr++ = da[i++];
			else
				*dr++ = db[j++];
	
		while( i<na ) *dr++ = da[i++];
		while( j<nb ) *dr++ = db[j++];

	}	
	/* unique :) */	
	if ( ARRSIZE(r) > 1 ) {
		tmp = dr = ARRPTR( r );
		while( tmp - ARRPTR( r ) < ARRSIZE(r) ) 
			if ( *tmp != *dr ) 
				*(++dr) = *tmp++;
			else 
				tmp++; 
	}	

	return resize_intArrayType(r, dr + 1 - ARRPTR(r) );
}


ArrayType * 
_int_inter ( ArrayType *a, ArrayType *b ) {
	ArrayType * r;
	int na , nb ;
	int *da, *db, *dr;
	int i,j;

	if ( ARRISNULL( a ) || ARRISNULL( b ) ) return NULL;

	PREPAREARR;

	r = new_intArrayType( min(na, nb) ); 
	dr = ARRPTR( r );
	
	i = j = 0;
	while( i<na && j<nb ) 
		if ( da[i] < db[j] )
			i++;
		else if ( da[i] == db[j] ) { 
			if ( i+j == 0 || ( i+j>0 && *(dr-1) != db[j] ) )  
				*dr++ = db[j];
			i++; j++;
		} else 
			j++;

	if ( (dr - ARRPTR(r)) == 0 ) {
		pfree( r );
		return NULL;
	} else 
		return resize_intArrayType(r, dr - ARRPTR(r) );
}

void
rt__int_size(ArrayType *a, float *size)
{
  if (a == (ArrayType *) NULL )
    *size = 0.0;
  else
    *size = (float)ARRSIZE( a );
  
  return;
}


/*****************************************************************************
 *                 Miscellaneous operators and functions
 *****************************************************************************/

int sort_intarray( const void *a, const void *b ) { 
        if ( *((int*)a) == *((int*)b) ) return 0;
        return ( *((int*)a) > *((int*)b) ) ? 1 : -1;
}

int bytesize_intarray (ArrayType * a) {
	if ( ! a ) return 0;
	return ARR_OVERHEAD( NDIM ) + sizeof(int)*ARRSIZE( a );
}

ArrayType * new_intArrayType( int num ) {
	ArrayType * r;
	int nbytes = ARR_OVERHEAD( NDIM ) + sizeof(int)*num;
	
	r = (ArrayType *) palloc( nbytes );
	if ( ! r )
		elog(ERROR, "Can't allocate memory for new array");
	MemSet(r, 0, nbytes);
	r->size = nbytes;
	r->ndim = NDIM;
	SET_LO_FLAG(false, r);
	*( (int*)ARR_DIMS(r) ) = num;
	*( (int*)ARR_LBOUND(r) ) = 1;
	
	return r;	
} 

ArrayType * resize_intArrayType( ArrayType * a, int num ) {
	int nbytes = ARR_OVERHEAD( NDIM ) + sizeof(int)*num;

	if ( num == ARRSIZE(a) ) return a;

	a = (ArrayType *) repalloc( a, nbytes );
	if ( ! a )
		elog(ERROR, "Can't reallocate memory for new array");
	
	a->size = nbytes;
	*( (int*)ARR_DIMS(a) ) = num; 
	
	return a;
}

ArrayType * copy_intArrayType( ArrayType * a ) {
	ArrayType * r;
	r = new_intArrayType( ARRSIZE(a) );
	memmove(r,a,bytesize_intarray(a));
	return r;
}

/* num for compressed key */
int internal_size (int *a, int len ) {
        int i,size=0;

        for(i=0;i<len;i+=2)
                if ( ! i || a[i] != a[i-1]  ) /* do not count repeated range */
                        size += a[i+1] - a[i] + 1;

        return size;
}

